/* asm-i386-gcc.h -- (C) Geoffrey Reynolds, November 2006.

   Inline i386 assember routines for GCC.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#ifndef _ASM_I386_GCC_H
#define _ASM_I386_GCC_H

#include "config.h"

/* Transparent unions are a GCC extension. When a function is declared to
   take transparant union arguments the argument is actually passed as the
   type of its first element. TU64_t types declared below are used to pass
   64 bit integers to low level functions that operate on their parts.
*/
typedef union
{
  uint64_t u;
  struct { uint32_t l, h; };
} __attribute__ ((transparent_union)) TU64_t;

typedef uint64_t v128_t attribute ((vector_size(16)));

#endif /* _ASM_I386_GCC_H */


/* This file may be included from a number of different headers. It should
   only define FEATURE if NEED_FEATURE is defined but HAVE_FEATURE is not.
*/


#if defined(NEED_mulmod64) && !defined(HAVE_mulmod64)
#define HAVE_mulmod64

/* All intermediate values are kept on the FPU stack, which gives correct
   results for all primes up to 2^62.
*/
#define MOD64_MAX_BITS 62

/* Old FPU rounding mode (bits 10 and 11) saved here before switching to
   round-to-zero mode. This is also used as a dummy dependancy for those
   functions which actually depend on the state of the FPU stack.
*/
extern uint16_t mod64_rnd;

/* a*b (mod p), where a,b < p < 2^62.
   Assumes %st(0) contains 1.0/p and FPU is in round-to-zero mode.
*/
#if 0
/* Use the code in mulmod-i386.S
 */
uint64_t mulmod64(uint64_t a, uint64_t b, uint64_t p) attribute ((pure));
#else
static inline uint64_t mulmod64(uint64_t a, uint64_t b, TU64_t p)
{
  uint32_t t1;
  uint32_t t2;
  uint64_t tmp;
  register uint64_t ret;

  asm("fildll  %5"                  "\n\t"
      "fildll  %4"                  "\n\t"
      "mov     %5, %%eax"           "\n\t"
      "mov     %5, %%ecx"           "\n\t"
      "mov     4+%5, %%ebx"         "\n\t"
      "mull    %4"                  "\n\t"
      "fmulp   %%st(0), %%st(1)"    "\n\t"
      "imul    %4, %%ebx"           "\n\t"
      "fmul    %%st(1), %%st(0)"    "\n\t"
      "imul    4+%4, %%ecx"         "\n\t"
      "mov     %%eax, %2"           "\n\t"
      "add     %%ebx, %%ecx"        "\n\t"
      "fistpll %1"                  "\n\t"
      "add     %%ecx, %%edx"        "\n\t"
      "mov     %%edx, %3"           "\n\t"
      "mov     %1, %%eax"           "\n\t"
      "mov     4+%1, %%edx"         "\n\t"
      "mov     %7, %%ebx"           "\n\t"
      "mov     %6, %%ecx"           "\n\t"
      "imul    %%eax, %%ebx"        "\n\t"
      "imul    %%edx, %%ecx"        "\n\t"
      "mull    %6"                  "\n\t"
      "add     %%ebx, %%ecx"        "\n\t"
      "mov     %3, %%ebx"           "\n\t"
      "add     %%ecx,%%edx"         "\n\t"
      "mov     %2, %%ecx"           "\n\t"
      "sub     %%eax, %%ecx"        "\n\t"
      "sbb     %%edx, %%ebx"        "\n\t"
      "mov     %%ecx, %%eax"        "\n\t"
      "mov     %%ebx, %%edx"        "\n\t"
      "sub     %6, %%ecx"           "\n\t"
      "sbb     %7, %%ebx"           "\n\t"
      "jl      0f"                  "\n\t"
      "mov     %%ecx, %%eax"        "\n\t"
      "mov     %%ebx, %%edx"        "\n\t"
      "0:"
      : "=&A" (ret), "=m" (tmp), "=m" (t1), "=m" (t2)
      : "m" (a), "m" (b), "rm" (p.l), "rm" (p.h), "m" (mod64_rnd)
      : "%ebx", "%ecx", "cc" );

  return ret;
}
#endif

/* Pushes 1.0/p onto the FPU stack and changes the FPU into round-to-zero
   mode. Saves the old rounding mode in mod64_rnd.
*/
static inline void mod64_init(uint64_t p)
{
  uint16_t tmp;

  asm volatile ("fildll  %2"               "\n\t"
                "fld1"                     "\n\t"
                "fnstcw  %0"               "\n\t"
                "mov     %0, %%ax"         "\n\t"
                "and     $3072, %%ax"      "\n\t"
                "orw     $3072, %0"        "\n\t"  /* Set bits 10,11 */
                "mov     %%ax, %1"         "\n\t"  /* Save old bits */
                "fldcw   %0"               "\n\t"
                "fdivp   %%st(0),%%st(1)"

                : "=m" (tmp), "+m" (mod64_rnd)
                : "m" (p)
                : "%eax", "cc" );
}

/* Same as mod64_init(p) but does not set FPU rounding, for nested use, e.g.
   mod64_init(p1)
   mod64_init0(p2)
   mod64_init0(p3)
   ...
   mod64_fini0()
   mod64_fini0()
   mod64_fini()
*/
static inline void mod64_init0(uint64_t p)
{
  asm volatile ("fildll  %1"               "\n\t"
                "fld1"                     "\n\t"
                "fdivp   %%st(0),%%st(1)"

                : "+m" (mod64_rnd) /* Dummy dependency */
                : "m" (p)
                : "cc" );
}

/* Pops 1.0/p off the FPU stack and restores the previous FPU rounding mode
   from mod64_rnd.
*/
static inline void mod64_fini(void)
{
  uint16_t tmp;

  asm volatile ("fstp    %%st(0)"          "\n\t"
                "fnstcw  %0"               "\n\t"
                "mov     %0, %%ax"         "\n\t"
                "and     $62463, %%ax"     "\n\t"  /* Clear bits 10,11 */
                "or      %1, %%ax"         "\n\t"  /* Restore old bits */
                "mov     %%ax, %0"         "\n\t"
                "fldcw   %0"

                : "=m" (tmp), "+m" (mod64_rnd)
                : 
                : "%eax", "cc" );
}

/* Same as mod64_fini() but does not restore rounding mode.
 */
static inline void mod64_fini0(void)
{
  asm volatile ("fstp    %%st(0)"

                : "+m" (mod64_rnd) /* Dummy dependency */
                : 
                : "cc" );
}


/* a*b (mod p), where a,b < p < 2^62.
   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define PRE2_MULMOD64(a,b,p) ({ \
  uint32_t r1, r2; \
  uint64_t tmp; \
  register uint64_t ret; \
  asm("fildll  %4"                  "\n\t" \
      "mov     %4, %%ebx"           "\n\t" \
      "mov     4+%4, %%ecx"         "\n\t" \
      "imul    %%edx, %%ebx"        "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "imul    %%eax, %%ecx"        "\n\t" \
      "fistpll %1"                  "\n\t" \
      "mull    %4"                  "\n\t" \
      "mov     %%eax, %2"           "\n\t" \
      "add     %%ecx, %%ebx"        "\n\t" \
      "add     %%ebx, %%edx"        "\n\t" \
      "mov     %1, %%eax"           "\n\t" \
      "mov     %%edx, %3"           "\n\t" \
      "mov     4+%1, %%edx"         "\n\t" \
      "mov     %7, %%ebx"           "\n\t" \
      "mov     %6, %%ecx"           "\n\t" \
      "imul    %%eax, %%ebx"        "\n\t" \
      "imul    %%edx, %%ecx"        "\n\t" \
      "mull    %6"                  "\n\t" \
      "add     %%ebx, %%ecx"        "\n\t" \
      "mov     %3, %%ebx"           "\n\t" \
      "add     %%ecx,%%edx"         "\n\t" \
      "mov     %2, %%ecx"           "\n\t" \
      "sub     %%eax, %%ecx"        "\n\t" \
      "sbb     %%edx, %%ebx"        "\n\t" \
      "mov     %%ecx, %%eax"        "\n\t" \
      "mov     %%ebx, %%edx"        "\n\t" \
      "sub     %6, %%ecx"           "\n\t" \
      "sbb     %7, %%ebx"           "\n\t" \
      "jl      0f"                  "\n\t" \
      "mov     %%ecx, %%eax"        "\n\t" \
      "mov     %%ebx, %%edx"        "\n"   \
      "0:" \
      : "=&A" (ret), "=m" (tmp), "=m" (r1), "=m" (r2) \
      : "m" ((uint64_t)a), "0" ((uint64_t)b), \
        "rm" (((TU64_t)p).l), "rm" (((TU64_t)p).h), "m" (mod64_rnd) \
      : "%ebx", "%ecx", "cc" ); \
  ret; })

/* Pushes b/p onto the FPU stack. Assumes %st(0) contains 1.0/p.
 */
#define PRE2_MULMOD64_INIT(b) \
  asm volatile ("fildll  %1\n\t" \
                "fmul    %%st(1),%%st(0)" \
                : "+m" (mod64_rnd) \
                : "m" ((uint64_t)b) \
                : "cc" )

/* Pops b/p off the FPU stack.
 */
#define PRE2_MULMOD64_FINI() \
  asm volatile ("fstp    %%st(0)" \
                : "+m" (mod64_rnd) \
                : \
                : "cc" )
#endif


#if defined(NEED_sqrmod64) && !defined(HAVE_sqrmod64)
#define HAVE_sqrmod64
extern uint16_t mod64_rnd;

/* b^2 (mod p), where b < p < 2^62.
   Assumes %st(0) contains 1.0/p and FPU is in round-to-zero mode.
*/
static inline uint64_t sqrmod64(uint64_t b, TU64_t p)
{
  uint32_t r1;
  uint32_t r2;
  uint64_t tmp;
  register uint64_t ret;

  asm("fildll  %4"                  "\n\t"
      "mov     %4, %%ecx"           "\n\t"
      "mov     %4, %%eax"           "\n\t"
      "mov     4+%4, %%ebx"         "\n\t"
      "mul     %%eax"               "\n\t"
      "fmul    %%st(0), %%st(0)"    "\n\t"
      "imul    %%ebx, %%ecx"        "\n\t"
      "fmul    %%st(1), %%st(0)"    "\n\t"
      "mov     %%eax, %2"           "\n\t"
      "fistpll %1"                  "\n\t"
      "lea     (%%edx,%%ecx,2), %%edx\n\t"
      "mov     %%edx, %3"           "\n\t"
      "mov     %1, %%eax"           "\n\t"
      "mov     4+%1, %%edx"         "\n\t"
      "mov     %6, %%ebx"           "\n\t"
      "mov     %5, %%ecx"           "\n\t"
      "imul    %%eax, %%ebx"        "\n\t"
      "imul    %%edx, %%ecx"        "\n\t"
      "mull    %5"                  "\n\t"
      "add     %%ebx, %%ecx"        "\n\t"
      "mov     %3, %%ebx"           "\n\t"
      "add     %%ecx,%%edx"         "\n\t"
      "mov     %2, %%ecx"           "\n\t"
      "sub     %%eax, %%ecx"        "\n\t"
      "sbb     %%edx, %%ebx"        "\n\t"
      "mov     %%ecx, %%eax"        "\n\t"
      "mov     %%ebx, %%edx"        "\n\t"
      "sub     %5, %%ecx"           "\n\t"
      "sbb     %6, %%ebx"           "\n\t"
      "jl      0f"                  "\n\t"
      "mov     %%ecx, %%eax"        "\n\t"
      "mov     %%ebx, %%edx"        "\n\t"
      "0:"
      : "=&A" (ret), "=m" (tmp), "=m" (r1), "=m" (r2)
      : "m" (b), "g" (p.l), "g" (p.h), "m" (mod64_rnd)
      : "%ebx", "%ecx", "cc" );

  return ret;
}
#endif


#if defined(NEED_powmod64) && !defined(HAVE_powmod64)
#define HAVE_powmod64
/*
  b^n (mod p)
*/
#ifdef __SSE2__
/* Use the code in powmod-sse2.S
 */
uint64_t powmod64_sse2(uint64_t b, uint64_t n, uint64_t p)
  attribute ((pure,regparm(2)));
#define powmod64(b,n,p) powmod64_sse2(n,b,p)
#else
/* Use the code in powmod-i386.S
 */
uint64_t powmod64(uint64_t b, uint64_t n, uint64_t p) attribute ((pure));
#endif
#endif


#if defined(NEED_rem64_32) && !defined(HAVE_rem64_32)
#define HAVE_rem64_32
static inline uint32_t rem64_32(uint64_t p64, uint32_t m32)
{
  register uint32_t quot, rem;

  asm ("divl    %3"
       : "=a" (quot), "=d" (rem)
       : "A" (p64), "rm" (m32)
       : "cc" );

  return rem;
}
#endif


#if defined(NEED_memset_fast32) && !defined(HAVE_memset_fast32)
#define HAVE_memset_fast32
/* store count copies of x starting at dst.
 */
static inline
void memset_fast32(uint_fast32_t *dst, uint_fast32_t x, uint_fast32_t count)
{
  register uint32_t tmp;
  register uint_fast32_t *ptr;

  asm ("cld\n\t"
       "rep; stosl"
       : "=c" (tmp), "=D" (ptr),
         "=m" (*(struct { uint_fast32_t dummy[count]; } *)dst)
       : "a" (x), "0" (count), "1" (dst)
       : "cc" );
}
#endif


#if defined(NEED_vec2_mulmod64) && !defined(HAVE_vec2_mulmod64)
#ifdef __SSE2__
#define HAVE_vec2_mulmod64

/* Given uint64_t X[4] and v128_t B={b,b}, P={p,p}:

   Assign X[i+2] <-- X[i] * b (mod p), for 0 <= i < 2.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/

#define VEC2_MULMOD64_NEXT2(X,B,P) { \
  asm("fildll  %4"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %0"                  "\n\t" \
      "fildll  %5"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
 \
      "movdqa  %4, %%xmm2"          "\n\t" \
      "movdqa  %2, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %2, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %2, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  %0, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %3, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %3, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %3, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %3, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %3, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      : "=m" ((X)[2]), "=m" ((X)[3]) \
      : "x" (B), "x" (P), "m" ((X)[0]), "m" ((X)[1]), "m" (mod64_rnd) \
      : "%xmm2", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}

/* Given uint64_t X[8] and v128_t B={b,b}, P={p,p}:

   Assign X[i+4] <-- X[i] * b (mod p), for 0 <= i < 4.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define VEC4_MULMOD64_NEXT4(X,B,P) { \
  asm("fildll  %6"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %0"                  "\n\t" \
      "fildll  %7"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
 \
      "movdqa  %6, %%xmm2"          "\n\t" \
      "movdqa  %4, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %4, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %4, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "fildll  %8"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %2"                  "\n\t" \
      "fildll  %9"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %3"                  "\n\t" \
 \
      "movdqa  %8, %%xmm3"          "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %4, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %4, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %0, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %5, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %5, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %5, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %5, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %5, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      "movdqa  %2, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %5, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %5, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %5, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %5, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %2" \
 \
      : "=m" ((X)[4]), "=m" ((X)[5]), "=m" ((X)[6]), "=m" ((X)[7]) \
      : "x" (B), "x" (P), "m" ((X)[0]), "m" ((X)[1]), "m" ((X)[2]), \
        "m" ((X)[3]), "m" (mod64_rnd) \
      : "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}

/* Given uint64_t X[16] and v128_t B={b,b}, P={p,p}:

   Assign X[i+8] <-- X[i] * b (mod p), for 0 <= i < 8.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define VEC8_MULMOD64_NEXT8(X,B,P) { \
  asm("fildll  %10"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %0"                  "\n\t" \
      "fildll  %11"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
 \
      "fildll  %12"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %2"                  "\n\t" \
      "fildll  %13"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %3"                  "\n\t" \
 \
      "fildll  %14"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %4"                  "\n\t" \
      "fildll  %15"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %5"                  "\n\t" \
 \
      "fildll  %16"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %6"                  "\n\t" \
      "fildll  %17"                 "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %7"                  "\n\t" \
 \
      "movdqa  %10, %%xmm2"         "\n\t" \
      "movdqa  %8, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %8, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %8, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  %12, %%xmm3"         "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %8, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %8, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %0, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %9, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %9, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %9, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %9, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      "movdqa  %2, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %9, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %9, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %9, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %2"          "\n\t" \
 \
 \
      "movdqa  %14, %%xmm2"          "\n\t" \
      "movdqa  %8, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %8, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %8, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  %16, %%xmm3"          "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %8, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %8, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %4, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %9, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %9, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %9, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %9, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %4"          "\n\t" \
 \
      "movdqa  %6, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %9, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %9, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %9, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %6" \
 \
      : "=m" ((X)[8]), "=m" ((X)[9]), "=m" ((X)[10]), "=m" ((X)[11]), \
        "=m" ((X)[12]), "=m" ((X)[13]), "=m" ((X)[14]), "=m" ((X)[15]) \
      : "x" (B), "x" (P), "m" ((X)[0]), "m" ((X)[1]), "m" ((X)[2]), \
        "m" ((X)[3]), "m" ((X)[4]), "m" ((X)[5]), "m" ((X)[6]), \
        "m" ((X)[7]), "m" (mod64_rnd) \
      : "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}

/* Given uint64_t X[32] and v128_t B={b,b}, P={p,p}:

   Assign X[i+16] <-- X[i] * b (mod p), for 0 <= i < 16.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define VEC16_MULMOD64_NEXT16(X,B,P) { \
  asm("fildll  -128+%0"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %0"                  "\n\t" \
      "fildll  -128+%1"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
 \
      "fildll  -128+%2"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %2"                  "\n\t" \
      "fildll  -128+%3"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %3"                  "\n\t" \
 \
      "fildll  -128+%4"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %4"                  "\n\t" \
      "fildll  -128+%5"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %5"                  "\n\t" \
 \
      "fildll  -128+%6"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %6"                  "\n\t" \
      "fildll  -128+%7"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %7"                  "\n\t" \
 \
      "fildll  -128+%8"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %8"                  "\n\t" \
      "fildll  -128+%9"             "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %9"                  "\n\t" \
 \
      "fildll  -128+%10"            "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %10"                 "\n\t" \
      "fildll  -128+%11"            "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %11"                 "\n\t" \
 \
      "movdqa  -128+%0, %%xmm2"     "\n\t" \
      "movdqa  %16, %%xmm5"         "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %16, %%xmm2"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "fildll  -128+%12"            "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %12"                 "\n\t" \
      "fildll  -128+%13"            "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %13"                 "\n\t" \
 \
      "movdqa  -128+%2, %%xmm3"     "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %16, %%xmm3"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "fildll  -128+%14"            "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %14"                 "\n\t" \
      "fildll  -128+%15"            "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %15"                 "\n\t" \
 \
      "movdqa  %0, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %17, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %17, %%xmm2"         "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %17, %%xmm4"         "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      "movdqa  %2, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %17, %%xmm3"         "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %17, %%xmm7"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %2"          "\n\t" \
 \
 \
      "movdqa  -128+%4, %%xmm2"     "\n\t" \
      "movdqa  %16, %%xmm5"         "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %16, %%xmm2"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  -128+%6, %%xmm3"     "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %16, %%xmm3"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %4, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %17, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %17, %%xmm2"         "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %17, %%xmm4"         "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %4"          "\n\t" \
 \
      "movdqa  %6, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %17, %%xmm3"         "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %17, %%xmm7"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %6"          "\n\t" \
 \
 \
 \
      "movdqa  -128+%8, %%xmm2"     "\n\t" \
      "movdqa  %16, %%xmm5"         "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %16, %%xmm2"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  -128+%10, %%xmm3"    "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %16, %%xmm3"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %8, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %17, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %17, %%xmm2"         "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %17, %%xmm4"         "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %8"          "\n\t" \
 \
      "movdqa  %10, %%xmm6"         "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %17, %%xmm3"         "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %17, %%xmm7"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %10"         "\n\t" \
 \
 \
      "movdqa  -128+%12, %%xmm2"    "\n\t" \
      "movdqa  %16, %%xmm5"         "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %16, %%xmm2"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  -128+%14, %%xmm3"    "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %16, %%xmm3"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %16, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %12, %%xmm6"         "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %17, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %17, %%xmm2"         "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %17, %%xmm4"         "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %12"         "\n\t" \
 \
      "movdqa  %14, %%xmm6"         "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm6"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %17, %%xmm3"         "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %17, %%xmm7"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %14" \
 \
      : "=m" ((X)[16]), "=m" ((X)[17]), "=m" ((X)[18]), "=m" ((X)[19]), \
        "=m" ((X)[20]), "=m" ((X)[21]), "=m" ((X)[22]), "=m" ((X)[23]), \
        "=m" ((X)[24]), "=m" ((X)[25]), "=m" ((X)[26]), "=m" ((X)[27]), \
        "=m" ((X)[28]), "=m" ((X)[29]), "=m" ((X)[30]), "=m" ((X)[31]) \
      : "x" (B), "x" (P), "m" (*(struct {uint64_t x[16];} *)(X)), \
        "m" (mod64_rnd) \
      : "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}


/* Given uint64_t X[4] and v128_t B={b,b}, P={p,p}:

   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 4.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define VEC4_MULMOD64_THIS4(X,B,P) { \
  asm("movdqa  %0, %%xmm2"          "\n\t" \
      "fildll  %0"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %0"                  "\n\t" \
      "fildll  %1"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
 \
      "movdqa  %4, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %4, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %4, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  %2, %%xmm3"          "\n\t" \
      "fildll  %2"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %2"                  "\n\t" \
      "fildll  %3"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %3"                  "\n\t" \
 \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %4, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %4, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %0, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %5, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %5, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %5, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %5, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %5, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      "movdqa  %2, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %5, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %5, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %5, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %5, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %2" \
 \
      : "+m" ((X)[0]), "+m" ((X)[1]), "+m" ((X)[2]), "+m" ((X)[3]) \
      : "x" (B), "x" (P), "m" (mod64_rnd) \
      : "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}


/* Given uint64_t X[8] and v128_t B={b,b}, P={p,p}:

   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 8.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define VEC8_MULMOD64_THIS8(X,B,P) { \
  uint64_t _T[8] attribute ((aligned(16))); \
 \
  asm("fildll  %0"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
      "fildll  8+%0"                "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %2"                  "\n\t" \
 \
      "fildll  16+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %3"                  "\n\t" \
      "fildll  24+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %4"                  "\n\t" \
 \
      "movdqa  %0, %%xmm2"          "\n\t" \
      "movdqa  %9, %%xmm5"         "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %9, %%xmm2"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "fildll  32+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %5"                  "\n\t" \
      "fildll  40+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %6"                  "\n\t" \
 \
      "movdqa  16+%0, %%xmm3"       "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %9, %%xmm3"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "fildll  48+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %7"                  "\n\t" \
      "fildll  56+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %8"                  "\n\t" \
 \
      "movdqa  %1, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %10, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %10, %%xmm4"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %10, %%xmm6"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %10, %%xmm2"         "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %10, %%xmm4"         "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      "movdqa  %3, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %10, %%xmm6"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %10, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %10, %%xmm3"         "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %10, %%xmm7"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, 16+%0"       "\n\t" \
 \
 \
      "movdqa  32+%0, %%xmm2"       "\n\t" \
      "movdqa  %9, %%xmm5"         "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %9, %%xmm2"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  48+%0, %%xmm3"       "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %9, %%xmm3"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %9, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %5, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %10, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %10, %%xmm4"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %10, %%xmm6"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %10, %%xmm2"         "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %10, %%xmm4"         "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, 32+%0"       "\n\t" \
 \
      "movdqa  %7, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %10, %%xmm6"         "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %10, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %10, %%xmm3"         "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %10, %%xmm7"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, 48+%0" \
 \
      : "+m" (*(struct {uint64_t x[8];} *)(X)), \
        "=m" (_T[0]), "=m" (_T[1]), "=m" (_T[2]), "=m" (_T[3]), \
        "=m" (_T[4]), "=m" (_T[5]), "=m" (_T[6]), "=m" (_T[7]) \
      : "x" (B), "x" (P), "m" (mod64_rnd) \
      : "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}

/* Given uint64_t X[16] and v128_t B={b,b}, P={p,p}:

   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 16.

   Assumes %st(0) contains b/p and FPU is in round-to-zero mode.
*/
#define VEC16_MULMOD64_THIS16(X,B,P) { \
  uint64_t _T[16] attribute ((aligned(16))); \
 \
  asm("fildll  %0"                  "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %1"                  "\n\t" \
      "fildll  8+%0"                "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %2"                "\n\t" \
 \
      "fildll  16+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %3"               "\n\t" \
      "fildll  24+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %4"               "\n\t" \
 \
      "fildll  32+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %5"               "\n\t" \
      "fildll  40+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %6"               "\n\t" \
 \
      "fildll  48+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %7"               "\n\t" \
      "fildll  56+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %8"               "\n\t" \
 \
      "fildll  64+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %9"               "\n\t" \
      "fildll  72+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %10"               "\n\t" \
 \
      "fildll  80+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %11"               "\n\t" \
      "fildll  88+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %12"               "\n\t" \
 \
      "movdqa  %0, %%xmm2"          "\n\t" \
      "movdqa  %17, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %17, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "fildll  96+%0"               "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %13"               "\n\t" \
      "fildll  104+%0"              "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %14"              "\n\t" \
 \
      "movdqa  16+%0, %%xmm3"       "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "fildll  112+%0"              "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %15"              "\n\t" \
      "fildll  120+%0"              "\n\t" \
      "fmul    %%st(1), %%st(0)"    "\n\t" \
      "fistpll %16"              "\n\t" \
 \
      "movdqa  %1, %%xmm6"          "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %18, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %18, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %18, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, %0"          "\n\t" \
 \
      "movdqa  %3, %%xmm6"       "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %18, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %18, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, 16+%0"       "\n\t" \
 \
 \
      "movdqa  32+%0, %%xmm2"       "\n\t" \
      "movdqa  %17, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %17, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  48+%0, %%xmm3"       "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %5, %%xmm6"       "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %18, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %18, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %18, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, 32+%0"       "\n\t" \
 \
      "movdqa  %7, %%xmm6"       "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %18, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %18, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, 48+%0"       "\n\t" \
 \
 \
 \
      "movdqa  64+%0, %%xmm2"       "\n\t" \
      "movdqa  %17, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %17, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  80+%0, %%xmm3"       "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %9, %%xmm6"       "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %18, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %18, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %18, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, 64+%0"       "\n\t" \
 \
      "movdqa  %11, %%xmm6"       "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %18, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %18, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, 80+%0"       "\n\t" \
 \
 \
      "movdqa  96+%0, %%xmm2"       "\n\t" \
      "movdqa  %17, %%xmm5"          "\n\t" \
      "movdqa  %%xmm2, %%xmm7"      "\n\t" \
      "movdqa  %%xmm2, %%xmm4"      "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "pmuludq %17, %%xmm2"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm2"      "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
 \
      "movdqa  112+%0, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %17, %%xmm3"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %17, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "paddq   %%xmm4, %%xmm3"      "\n\t" \
 \
      "movdqa  %13, %%xmm6"       "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "movdqa  %%xmm6, %%xmm4"      "\n\t" \
      "movdqa  %18, %%xmm5"          "\n\t" \
      "psrlq   $32, %%xmm4"         "\n\t" \
      "psrlq   $32, %%xmm5"         "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm4, %%xmm4"      "\n\t" \
      "psubq   %%xmm6, %%xmm2"      "\n\t" \
      "psubq   %18, %%xmm2"          "\n\t" \
      "pcmpgtd %%xmm2, %%xmm4"      "\n\t" \
      "pshufd  $0xF5,%%xmm4,%%xmm4" "\n\t" \
      "pand    %18, %%xmm4"          "\n\t" \
      "paddq   %%xmm4, %%xmm2"      "\n\t" \
      "movdqa  %%xmm2, 96+%0"       "\n\t" \
 \
      "movdqa  %15, %%xmm6"      "\n\t" \
      "movdqa  %%xmm6, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm4" "\n\t" \
      "pmuludq %18, %%xmm6"          "\n\t" \
      "pmuludq %%xmm5, %%xmm7"      "\n\t" \
      "pmuludq %18, %%xmm4"          "\n\t" \
      "psllq   $32, %%xmm7"         "\n\t" \
      "psllq   $32, %%xmm4"         "\n\t" \
      "paddq   %%xmm7, %%xmm6"      "\n\t" \
      "paddq   %%xmm4, %%xmm6"      "\n\t" \
 \
      "pxor    %%xmm7, %%xmm7"      "\n\t" \
      "psubq   %%xmm6, %%xmm3"      "\n\t" \
      "psubq   %18, %%xmm3"          "\n\t" \
      "pcmpgtd %%xmm3, %%xmm7"      "\n\t" \
      "pshufd  $0xF5,%%xmm7,%%xmm7" "\n\t" \
      "pand    %18, %%xmm7"          "\n\t" \
      "paddq   %%xmm7, %%xmm3"      "\n\t" \
      "movdqa  %%xmm3, 112+%0"      "\n\t" \
 \
      : "+m" (*(struct {uint64_t x[16];} *)(X)), \
        "=m" (_T[0]), "=m" (_T[1]), "=m" (_T[2]), "=m" (_T[3]), \
        "=m" (_T[4]), "=m" (_T[5]), "=m" (_T[6]), "=m" (_T[7]), \
        "=m" (_T[8]), "=m" (_T[9]), "=m" (_T[10]), "=m" (_T[11]), \
        "=m" (_T[12]), "=m" (_T[13]), "=m" (_T[14]), "=m" (_T[15]) \
      : "x" (B), "x" (P), "m" (mod64_rnd) \
      : "%xmm2", "%xmm3", "%xmm4", "%xmm5", "%xmm6", "%xmm7" ); \
}

/* Return x, where:
   x=0x0000 (0)      if U[0]!=V[0] and U[1]!=V[1]
   x=0x00FF (255)    if U[0]==V[0] and U[1]!=V[1]
   x=0xFF00 (65280)  if U[0]!=V[0] and U[1]==V[1]
   x=0xFFFF (65535)  if U[0]==V[0] and U[1]==V[1]
 */
static inline uint32_t vec2_cmpeq(v128_t U, v128_t V)
{
  register uint32_t ret;
  register v128_t tmp1, tmp2;

  asm ("pcmpeqd  %3, %1"           "\n\t"
       "pshufd   $0xB1, %1, %2"    "\n\t"
       "pand     %1, %2"           "\n\t"
       "pmovmskb %2, %0"

       : "=r" (ret), "=x" (tmp1), "=x" (tmp2)
       : "x" (U), "1" (V) );

  return ret;
}

/* Set lower quadword to x, upper quadword to y. This is needed because of a
   bug in GCC 4.1 that causes X = (v128_t){x,y} to load both x and y into
   the low quadword of X!
 */
static inline v128_t vec2_set(uint64_t x, uint64_t y)
{
  v128_t ret;

  asm ("movq       %1, %0"    "\n\t"
       "movhps     %2, %0"

       : "=&x" (ret)
       : "m" (x), "m" (y) );

  return ret;
}

/* Set both quadwords to x. This should be the same as X = (v128_t){x,x},
   but avoids a potential bug in GCC 4.1.
 */
static inline v128_t vec2_set2(uint64_t x)
{
  v128_t ret;

  asm ("movq       %1, %0"    "\n\t"
       "punpcklqdq %0, %0"

       : "=x" (ret)
       : "m" (x) );

  return ret;
}
#endif
#endif
