/* powmod-k8.S -- (C) Geoffrey Reynolds, February 2007.

   uint64_t powmod64_k8(uint64_t b, uint64_t n, uint64_t p, double invp);
     Returns b^n (mod p), where 0 <= b < p < 2^52.
     Assumes invp = 1.0/p
     Assumes SSE rounding mode is round-to-zero.

   uint64_t powmod64_k8_fpu(uint64_t b, uint64_t n, uint64_t p);
     Returns b^n (mod p), where 0 <= b < p < 2^62.
     Assumes %st(0) contains 1.0/p
     Assumes FPU rounding mode is round-to-zero.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include "config.h"

#if USE_FPU_MULMOD
#if NEED_UNDERSCORE
#define powmod64_k8_fpu _powmod64_k8_fpu
#endif

	.text
	.p2align 4
	.globl powmod64_k8_fpu

powmod64_k8_fpu:
	mov	$1, %eax
	/* %rax = a = 1, %rdi = b, %rsi = n, %rdx = p, %xmm0 = 1.0/p */
tst:
	test	%rsi, %rsi
	jz	out
	mov	%rdi, -8(%rsp)
	mov	%rax, -16(%rsp)
	test	$1, %sil
	fildll  -8(%rsp)
	jz	sqr
mul:
	fildll  -16(%rsp)
	imul    %rdi, %rax
	fmulp   %st(0), %st(1)
	fmul    %st(1), %st(0)
	fistpll -16(%rsp)
	mov     -16(%rsp), %rcx
	imul    %rdx, %rcx
	sub     %rcx, %rax
	mov     %rax, %rcx
	sub     %rdx, %rax
	cmovl   %rcx, %rax
sqr:
	imul    %rdi, %rdi
	fmul    %st(0), %st(0)
	fmul    %st(1), %st(0)
	fistpll -16(%rsp)
	shr	%rsi
	mov     -16(%rsp), %rcx
	imul    %rdx, %rcx
	sub     %rcx, %rdi
	mov     %rdi, %rcx
	sub     %rdx, %rdi
	cmovl   %rcx, %rdi
	jmp	tst
out:
	ret

#else /* !USE_FPU_MULMOD */

#if NEED_UNDERSCORE
#define powmod64_k8 _powmod64_k8
#endif

	.text
	.p2align 4
	.globl powmod64_k8

powmod64_k8:
	mov	$1, %eax
	/* %rax = a = 1, %rdi = b, %rsi = n, %rdx = p, %xmm0 = 1.0/p */
tst:
	test	%rsi, %rsi
	jz	out
	test	$1, %sil
	cvtsi2sd %rdi, %xmm1
	jz	sqr
mul:
	cvtsi2sd %rax, %xmm2
	imul	%rdi, %rax
	mulsd	%xmm1, %xmm2
	mulsd	%xmm0, %xmm2
	cvtsd2si %xmm2, %rcx
	imul	%rdx, %rcx
	sub	%rcx, %rax
	mov	%rax, %rcx
	sub	%rdx, %rax
	cmovl	%rcx, %rax
sqr:
	imul	%rdi, %rdi
	mulsd	%xmm1, %xmm1
	mulsd	%xmm0, %xmm1
	cvtsd2si %xmm1, %rcx
	shr	%rsi
	imul	%rdx, %rcx
	sub	%rcx, %rdi
	mov	%rdi, %rcx
	sub	%rdx, %rdi
	cmovl	%rcx, %rdi
	jmp	tst
out:
	ret

#endif
