/* powmod-sse2.S -- (C) Geoffrey Reynolds, January 2007.

   uint64_t powmod64_sse2(uint64_t n, uint64_t b, uint64_t p);

   Returns b^n (mod p), where 0 <= b < p < 2^62.
   Assumes %st(0) contains 1.0/p.
   Assumes FPU rounding mode is round-to-zero.
   Assumes stack is 16-aligned.
   First argument is passed in %edx:%eax [function attribute regparm(2)].

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include "config.h"

#if NEED_UNDERSCORE
#define powmod64_sse2 _powmod64_sse2
#endif

	.text
	.p2align 4
	.globl	powmod64_sse2

powmod64_sse2:
	movq	12(%esp), %xmm0
	punpcklqdq %xmm0, %xmm0
	movl	$1, 12(%esp)
	movl	$0, 16(%esp)
	movdqa	%xmm0, %xmm6
	psrlq	$32, %xmm6
	movdqa	4(%esp), %xmm3
	jmp	test

	.p2align 4
sqr:
	pshufd	$0x4E, %xmm3, %xmm2	/* Save upper quadword */
	punpcklqdq %xmm3, %xmm3
	pshufd	$0xBE, %xmm2, %xmm1
	pmuludq	%xmm1, %xmm3
	pshufd	$0xEE, %xmm3, %xmm1
	fmulp	%st(0), %st(1)
	fistpll	4(%esp)
	psllq	$33, %xmm1
	paddq	%xmm1, %xmm3

	movq	4(%esp), %xmm1
	pshufd	$0x14, %xmm0, %xmm4
	pshufd	$0x41, %xmm1, %xmm5
	pmuludq	%xmm0, %xmm1
	pmuludq	%xmm5, %xmm4
	psllq	$32, %xmm4
	paddq	%xmm4, %xmm1
	psrldq	$8, %xmm4
	paddq	%xmm4, %xmm1

	pxor	%xmm5, %xmm5
	psubq	%xmm1, %xmm3
	psubq	%xmm0, %xmm3
	pcmpgtd	%xmm3, %xmm5
	pshufd	$0xF5, %xmm5, %xmm5
	pand	%xmm0, %xmm5
	paddq	%xmm5, %xmm3
	movq	%xmm3, 4(%esp)
	punpcklqdq %xmm2, %xmm3		/* Restore upper quadword */
shift:
	shrd	$1, %edx, %eax
	shr	%edx
test:
	mov	%eax, %ecx
	or	%edx, %ecx
	jz	out
	fildll  4(%esp)
	fld	%st(1)
	test	$1, %al
	fmul    %st(1), %st(0)
	jz	sqr

	movdqa	%xmm3, %xmm2
	fildll	12(%esp)
	pshufd	$0x44, %xmm3, %xmm1
	movdqa	%xmm2, %xmm4
	movdqa	%xmm1, %xmm5
	fmul	%st(1), %st(0)
	fistpll	12(%esp)
	fmulp	%st(0), %st(1)
	fistpll	4(%esp)
	pmuludq	%xmm1, %xmm3
	psrlq	$32, %xmm5
	psrlq	$32, %xmm4
	pmuludq	%xmm5, %xmm2
	pmuludq	%xmm1, %xmm4
	psllq	$32, %xmm2
	psllq	$32, %xmm4
	paddq	%xmm2, %xmm3
	paddq	%xmm4, %xmm3

	movdqa	4(%esp), %xmm1
	movdqa	%xmm1, %xmm2
	pmuludq	%xmm0, %xmm1
	movdqa	%xmm2, %xmm4
	psrlq	$32, %xmm4
	pmuludq	%xmm6, %xmm2
	pmuludq	%xmm0, %xmm4
	psllq	$32, %xmm2
	psllq	$32, %xmm4
	paddq	%xmm2, %xmm1
	paddq	%xmm4, %xmm1

	pxor	%xmm5, %xmm5
	psubq	%xmm1, %xmm3
	psubq	%xmm0, %xmm3
	pcmpgtd	%xmm3, %xmm5
	pshufd	$0xF5, %xmm5, %xmm5
	pand	%xmm0, %xmm5
	paddq	%xmm5, %xmm3
	movdqa	%xmm3, 4(%esp)

	jmp	shift
out:
	mov	12(%esp), %eax
	mov	16(%esp), %edx
	ret
